from .differentiation import Derivative, register
import numpy as np
from numpy.polynomial.polynomial import polyfit


@register("finite_difference")
class FiniteDifference(Derivative):
    def __init__(self, k, **kwargs):
        """
        Compute the symmetric numerical derivative of equally-spaced data using the Taylor series. Derivatives at the
        boundaries are computed with the available reduction of the window or first order finite difference.

        A simple first order finite difference scheme in the spirit of this code is available in numpy.gradient.

        Args:
            k (int): Interpolate the data with a polynomial through the 2k+1 points x[i-k], ..., x[i], ... x[i+k]
            **kwargs: Optional keyword arguments.

        Keyword Args:
            periodic (bool): If True, wrap the data. If False, compute edges separately. Default False.
        """
        self.k = k
        self.periodic = kwargs.get('periodic', False)

    @staticmethod
    def _symfd(i, x, dt, k):
        res = []
        # Construct each coefficient from the Taylor series.
        coefficient = 1
        for j in range(1, k + 1):
            coefficient *= (k - j + 1) / (k + j)
            alpha_j = 2 * np.power(-1, j + 1) * coefficient
            res.append(alpha_j * (x[(i + j) % len(x)] - x[(i - j) % len(x)]) / (2 * j) / dt)
        return np.sum(res)

    def compute(self, t, x, i):
        dt = t[1] - t[0]

        if not self.periodic:
            # Check for boundaries.
            if i == 0:
                return (x[i+1] - x[i])/dt
            elif i == len(x) - 1:
                return (x[i] - x[i-1])/dt

            # Check for overflow.
            left = i - self.k
            right = (len(x) - 1) - (i + self.k)
            overflow = min(left, right)
            if overflow < 0:
                return self._symfd(i, x, dt, self.k + overflow)

        # Default behavior (periodic data, symmetric derivative)
        return self._symfd(i, x, dt, self.k)


@register("savitzky_golay")
class SavitzkyGolay(Derivative):
    def __init__(self, order, left, right, **kwargs):
        """ Compute the numerical derivative by first finding the best (least-squares) polynomial of order m < 2k+1
        using the 2k+1 points in the neighborhood [t-left, t+right]. The derivative is computed  from the coefficients
        of the polynomial. The default edge behavior is to truncate the window at the offending edge.

        A simple symmetric version of the Savitzky-Galoy filter is available as scipy.signal.savgol_filter.

        Args:
            left (float): left edge of the window is t-left
            right (float): right edge of the window is t+right
            order (int):  order of polynomial (m < points in window)
            **kwargs: Optional keyword arguments.

        Keyword Args:
            use_iwindow (bool): Whether to use an indexed window. If True, left and right act as indicies for t instead
                of as lengths in units of t. Default False.
        """
        self.left = left
        self.right = right
        self.use_iwindow = kwargs.get('use_iwindow', False)
        self.order = order

    def _trunc_window(self, t, i):
        ileft = np.argmin(np.abs(t - (t[i] - self.left)))
        iright = np.argmin(np.abs(t - (t[i] + self.right)))
        return [ileft, iright]

    def compute(self, t, x, i):
        # Default edge behavior is to truncate the window.
        if self.use_iwindow:
            i_l, i_r = max(0., i - self.left), min(i + self.right, len(t) - 1)
        else:
            i_l, i_r = self._trunc_window(t, i)

        # Construct a polynomial in t using least squares regression.
        tfit = t[i_l:i_r]
        xfit = x[i_l:i_r]
        # Can raise RankWarning if order exceeds points in the window.
        w = polyfit(tfit, xfit, self.order)

        # Compute the derivative from the polyfit coefficients.
        return np.sum([j * w[j] * np.power(t[i], j - 1) for j in range(1, self.order + 1)])



import abc
import numpy as np


methods = {}
default = ["finite_difference", {"k": 1}]


def register(name=""):
    def inner(f):
        n = name or f.__name__
        methods[n] = f
        return f
    return inner


def dxdt(x, t, kind=None, axis=1, **kwargs):
    """
    Compute the derivative of x with respect to t along axis using the numerical derivative specified by "kind".
    This is the functional interface of the Derivative class.

    Args:
        x (:obj:`ndarray` of float): Ordered measurement values.
        t (:obj:`ndarray` of float): Ordered measurement times.
        kind (string): Derivative method name.
            Built in kinds:
            - finite_difference. required kwargs: k (window size).
            - savitzky_golay. required kwargs: order, left, right. optional kwargs: use_iwindow.
            - spectral. required kwargs: None. optional kwargs: filter (frequency filter function).
            - spline. required kwargs: s (smoothing).
            - trend_filtered. required kwargs: order, alpha (regularization).
        axis ({0,1}). axis of x along which to differentiate. default 1.
        **kwargs: Keyword arguments for the derivative method "kind".

    Returns:
        :obj:`ndarray` of float: Returns dx/dt along axis.
    """
    if kind is None:
        method = methods.get(default[0])
        return method(**default[1]).d(x, t)
    else:
        method = methods.get(kind)
        return method(**kwargs).d(x, t, axis=axis)


class Derivative(abc.ABC):
    """ Interface for computing numerical derivatives. """

    @abc.abstractmethod
    def compute(self, t, x, i):
        """ Compute the derivative of x with respect to t at the index i of x, (dx/dt)[i].

        Computation of a derivative should fail explicitely if the implementation is unable to compute a derivative at
        the desired index. Used for global differentiation methods, for example.

        Args:
            t (:obj:`ndarray` of float):  Ordered measurement times.
            x (:obj:`ndarray` of float):  Ordered measurement values.
            i (int): Index i at which to compute (dx/dt)[i]

        Returns:
            float: (dx/dt)[i]
        """

    def compute_for(self, t, x, indices):
        """
        Compute derivative (dx/dt)[i] for i in indices. Overload this if
        desiring a more efficient computation over a list of indices.

        Args:
            t (:obj:`ndarray` of float): Ordered measurement times.
            x (:obj:`ndarray` of float): Ordered measurement values.
            indices (:obj:`ndarray` of int): Indices i at which to compute (dx/dt)[i]

        Returns:
            Generator[float]: yields (dx/dt)[i] for i in indices
        """
        for i in indices:
            yield self.compute(t, x, i)

    def d(self, X, t, axis=1):
        """
        Compute the derivative of measurements X taken at times t.
        
        Args:
            t (:obj:`ndarray` of float): Ordered measurement times.
            X  (:obj:`ndarray` of float): Ordered measurement values.
            axis ({0,1}). axis of X along which to differentiate. default 1.

        Returns:
            :obj:`ndarray` of float: Returns dX/dt along axis.

        Raises:
            ValueError: Requires that X.shape[axis] equals len(t). If X is flat, requires that len(X) equals len(t).
        """
        # Cast
        X = np.array(X)
        flat = False
        # Check shape and axis
        if len(X.shape) == 1:
            X = X.reshape(1, -1)
            flat = True
        elif len(X.shape) == 2:
            if axis == 0:
                X = X.T
            elif axis == 1:
                pass
            else:
                raise ValueError("Invalid axis.")
        else:
            raise ValueError("Invalid shape of X.")

        if X.shape[1] != len(t):
            raise ValueError("Desired X axis size does not match t size.")

        # Differentiate
        dX = np.array([list(self.compute_for(t, x, np.arange(len(t)))) for x in X])
        if flat:
            return dX.flatten()
        else:
            return dX if axis == 1 else dX.T



from .differentiation import Derivative, register
from .utils import deriv, integ

import numpy as np
from numpy.linalg import inv
from scipy import interpolate
from scipy.linalg import null_space
from sklearn.linear_model import Lasso


@register("spectral")
class Spectral(Derivative):
    def __init__(self, **kwargs):
        """
        Compute the numerical derivative by first computing the FFT. In Fourier space, derivatives are multiplication
        by i*phase; compute the IFFT after.
        Args:
            **kwargs: Optional keyword arguments.

        Keyword Args:
            filter: Optional. A function that takes in frequencies and outputs weights to scale the coefficient at
                the input frequency in Fourier space. Input frequencies are the discrete fourier transform sample
                frequencies associated with the domain variable. Look into python signal processing resources in
                scipy.signal for common filters.

        """
        # Filter function. Default: Identity filter
        self.filter = kwargs.get('filter', np.vectorize(lambda f: 1))
        self._x_hat = None
        self._freq = None

    def _dglobal(self, t, x):
        self._x_hat = np.fft.fft(x)
        self._freq = np.fft.fftfreq(t.size, d=(t[1] - t[0]))

    def compute(self, t, x, i):
        return next(self.compute_for(t, x, [i]))

    def compute_for(self, t, x, indices):
        self._dglobal(t, x)
        res = np.fft.ifft(1j * 2 * np.pi * self._freq * self.filter(self._freq) * self._x_hat).real
        for i in indices:
            yield res[i]


@register("spline")
class Spline(Derivative):
    def __init__(self, s, **kwargs):
        """ Compute the numerical derivative of data x using a (Cubic) spline (the Cubic spline is a natural choice as
         it minimizes the curvature of the fit). Compute the derivative from the form of the known Spline polynomials.

        Args:
            s (float): Amount of smoothing

        Keyword Args:
            order (int): Default is cubic spline (3).
            periodic (bool): Default False.
        """
        self.smoothing = s
        self.order = kwargs.get('order', 3)
        self.periodic = kwargs.get('periodic', False)

        self._t = None
        self._x = None
        self._spl = None

    def _global(self, t, x):
        self._loaded = True
        self._t = t
        self._x = x
        # returns (knots, coefficients, order)
        self._spl = interpolate.splrep(self._t, self._x, k=self.order, s=self.smoothing, per=self.periodic)

    def compute(self, t, x, i):
        self._global(t, x)
        return interpolate.splev(self._t[i], self._spl, der=1)

    def compute_for(self, t, x, indices):
        self._global(t, x)
        for i in indices:
            yield interpolate.splev(self._t[i], self._spl, der=1)


@register("trend_filtered")
class TrendFiltered(Derivative):
    def __init__(self, order, alpha, **kwargs):
        """ Compute the numerical derivative using Total Squared Variations,

            min_u (1/2)||A u - (x-x_0)||_2^2 + \alpha ||D^(order+1) u||_1

        where A is the linear integral operator, and D is the linear derivative operator. The vector u finds a
        global derivative that is a piecewise function made up of polynomials of the provided order.

        If order=0, this is equivalent to the total-variational derivative.

        Args:
            order (int): Order of the inner LASSO derivative
            alpha (float): Regularization hyper-parameter for the LASSO problem.
            **kwargs: Keyword arguments for the sklearn LASSO function.
        """
        self.dorder = order + 1
        self.alpha = alpha
        self.kwargs = kwargs

        self._t = None
        self._x = None
        self._model = None
        self._res = None

    @staticmethod
    def _nullsp(order, diffop):
        #  Analytic solutions to the null space are polynomials.
        n = diffop.shape[1]
        if order == 1:
            nullsp = np.ones([1, n]) / np.sqrt(n)
        elif order == 2:
            v1 = np.ones(n) / np.sqrt(n)
            v2 = np.linspace(1, -1, diffop.shape[1])
            nullsp = np.vstack([v1, v2 / np.sqrt(v2 ** 2)])
        elif order == 3:
            v1 = np.ones(n) / np.sqrt(n)
            v2 = np.linspace(1, -1, diffop.shape[1])
            v3 = -v2 ** 2 + np.sum(v2 ** 2) / n
            nullsp = np.vstack([v1, v2 / np.sqrt(v2 ** 2), v3 / np.sqrt(v3 ** 2)])
        else:
            nullsp = null_space(diffop).T  # Potential bottleneck! (Requires SVDs)
        return nullsp

    def _global(self, t, x):
        self._t = t
        self._x = x

        # I: Integrals and derivatives
        n = len(self._t)
        dt = self._t[1] - self._t[0]
        # Note: Penalize jump count with + np.identity(n)[:-order, :]
        arrD = deriv(n, dt, order=self.dorder)
        arrI = integ(n, dt)

        # II: Append null space of derivative
        Dtilde = np.vstack([arrD, self._nullsp(self.dorder, arrD)])

        # III: Compute workers X1 and X2
        invDtilde = inv(Dtilde)
        XDtilde = arrI.dot(invDtilde)
        X1, X2 = XDtilde[:, :-1], XDtilde[:, -1:]

        # IV: Compute projectors
        # Note: regarding inv, X2.T@X2 is always small
        X2_proj = inv(X2.T.dot(X2)).dot(X2.T)
        proj = np.identity(n) - X2.dot(X2_proj)

        # V: LASSO parameters; fit
        A = proj.dot(X1)
        b = proj.dot(self._x - self._x[0])
        self._model = Lasso(alpha=self.alpha / n, fit_intercept=False, **self.kwargs)
        self._model.fit(A, b)

        # VI: Restore desired variables
        theta1_hat = self._model.coef_
        theta2_hat = X2_proj.dot(self._x - self._x[0] - X1.dot(theta1_hat))
        self._res = invDtilde.dot(np.hstack([theta1_hat, theta2_hat]))

    def compute(self, t, x, i):
        self._global(t, x)
        return self._res[i]

    def compute_for(self, t, x, indices):
        self._global(t, x)
        for i in indices:
            yield self._res[i]



import numpy as np


def deriv(n, dx, order=1):
    """ Matrix derivative. Equi-spaced derivative via forward finite differences, mapping R^n into R^(n-order).

    Args:
        n (int): Number of data points
        dx (float): Width of x[k+1] - x[k]
        order (int): Order of finite difference method to use. Default 1.

    Return:
        (:obj:`ndarray` of float): An n-by-(n-order) matrix derivative.

    Raises:
        ValueError: Requires n >= 2 and order in {1,2,3}.
    """
    if n < 2:
        raise ValueError('Bad length of {}'.format(n))

    if order == 1:
        method = [-1, 1]
    elif order == 2:
        method = [-1, 2, -1]
    elif order == 3:
        method = [-1, 3, -3, 1]
    else:
        raise ValueError('Order {} unimplemented.'.format(order))
    M = [method + [0] * (n - len(method))]
    for row in range(n - len(method)):
        M.append([0] * (row + 1) + method + [0] * (n - len(method) - row - 1))
    return np.array(M) / dx ** order


def integ(n, dx=1):
    """ Equi-spaced anti-derivative via the trapezoid rule mapping R^n to R^n.

    Args:
        n (int): Number of data points
        dx (float): Width of x[k+1] - x[k]

    Return:
        (:obj:`ndarray` of float): An n-by-n matrix integral.
    Raises:
        ValueError: Requires n in positive integers.
    """
    if n == 1:
        return np.array([])
    elif n == 2:
        return np.array([[0, 0],
                         [1, 1]]) * dx / 2
    elif n > 2:
        M = [[0] * n]
        for row in range(0, n - 1):
            M.append([1] + [2] * row + [1] + [0] * (n - row - 2))
        return np.array(M) * dx / 2
    else:
        raise ValueError('Bad size of {}'.format(n))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['primelab']

package_data = \
{'': ['*']}

install_requires = \
['numpy>=1.18.3,<2.0.0', 'scipy>=1.4.1,<2.0.0', 'sklearn>=0.0,<0.1']

extras_require = \
{'docs': ['sphinx>=3.0.2,<4.0.0',
          'nbsphinx>=0.6.1,<0.7.0',
          'ipykernel>=5.2.1,<6.0.0',
          'jupyter_client>=6.1.3,<7.0.0']}

setup_kwargs = {
    'name': 'primelab',
    'version': '0.1.1',
    'description': 'Numerical differentiation in python.',
    'long_description': '.. image:: https://readthedocs.org/projects/prime/badge/?version=latest\n   :target: https://prime.readthedocs.io/en/latest/?badge=latest\n   :alt: Documentation Status\n  \n.. image:: https://img.shields.io/badge/License-MIT-blue.svg\n   :target: https://lbesson.mit-license.org/\n   :alt: MIT License\n \nNumerical differentiation methods for python, including:\n\n1. Symmetric finite difference schemes using arbitrary window size. \n\n2. Savitzky-Galoy derivatives of any polynomial order with independent left and right window parameters.\n\n3. Spectral derivatives with optional filter.\n\n4. Spline derivatives of any order.\n\n5. Polynomial-trend-filtered derivatives generalizing methods like total variational derivatives. \n\nThese examples are intended to survey some common differentiation methods. The goal of this package is to bind these common differentiation methods to an easily implemented differentiation interface to encourage user adaptation.\n\nUsage:\n\n.. code-block:: python\n\n    from primelab import dxdt\n    import numpy as np\n\n    t = np.linspace(0,2*np.pi,50)\n    x = np.sin(x)\n\n    # Finite differences with central differencing using 3 points.\n    result1 = dxdt(x, t, kind="finite_difference", k=1)\n\n    # Savitzky-Golay using cubic polynomials to fit in a centered window of length 1\n    result2 = dxdt(x, t, kind="savitzky_golay", left=.5, right=.5, order=3)\n\n    # Spectral derivative\n    result3 = dxdt(x, t, kind="spectral")\n\n    # Spline derivative with smoothing set to 0.01\n    result4 = dxdt(x, t, kind="spline", s=1e-2)\n\n    # Total variational derivative with regularization set to 0.01\n    result5 = dxdt(x, t, kind="trend_filtered", order=1, alpha=1e-2)\n\nProject references:\n\n[1] Numerical differentiation of experimental data: local versus global methods- K. Ahnert and M. Abel  \n\n[2] Numerical Differentiation of Noisy, Nonsmooth Data- Rick Chartrand  \n\n[3] The Solution Path of the Generalized LASSO- R.J. Tibshirani and J. Taylor\n',
    'author': 'Andy Goldschmidt',
    'author_email': 'andygold@uw.edu',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/andgoldschmidt/primelab',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
